
// This requires a sufficiently new dbghelp.dll
// undefining this will disable stack traces
#ifndef NDEBUG
#define USE_DBGHELP


class StackTrace
{
  CONTEXT cx;
  STACKFRAME64 sf;
  DWORD dLineOffset;
  IMAGEHLP_MODULE64 module;
  union
  {
    IMAGEHLP_SYMBOL64 sym;
    DWORD symbol_buffer[1024];
  };
  DWORD64 dOffset;
  IMAGEHLP_LINE64 line;

  const HANDLE hProcess;
  
  bool bDeinit;

public:
  StackTrace(CONTEXT &cx_);
  ~StackTrace() { if (bDeinit) Deinit(); }

  inline bool Init(void)
  { return (bDeinit = (!!SymInitialize(hProcess, NULL, TRUE))); }
  inline void Deinit(void) { bDeinit = (!SymCleanup(hProcess)); }

  inline bool Walk(void)
  {
    return !!StackWalk64(IMAGE_FILE_MACHINE_I386,
                hProcess,
                GetCurrentThread(),
                &sf,
                &cx,
                NULL /*ReadProcessMemoryProc64*/,
                SymFunctionTableAccess64,
                SymGetModuleBase64,
                NULL);
  }

  inline bool GetModule(void)
  {
    return !!SymGetModuleInfo64(hProcess, sf.AddrPC.Offset, &module);
  }

  inline bool GetSymbol(void)
  {
    sym.Name[0] = 0;
    dOffset = 0;
    return !!SymGetSymFromAddr64(
        hProcess,
        sf.AddrPC.Offset,
        &dOffset,
        &sym);
  }

  inline bool GetLine(void)
  {
    line.FileName = "";
    line.LineNumber = 0;
    dLineOffset = 0;
    return !!SymGetLineFromAddr64(
        hProcess,
        sf.AddrPC.Offset,
        &dLineOffset,
        &line);
  }

  void GetCurrentPosition(char *buffer, unsigned long size);
};

#endif
