//#undef NDEBUG
#include "module.h"

#ifndef NO_SCRIPT_ACTIONS

/* no point in specifying len separately, since there's never subitems */ 
#define DEF_SAN(name) { #name, sizeof(#name)-1 },

s_ScriptActionName ScriptActionStrings[ScriptActionModuleCount+1] =
{
  /* These MUST be in the same order as the */ 
  /* entries in the ScriptActionTypes enum  */ 

  /* module or window */ 
  DEF_SAN(Create)
  DEF_SAN(Destroy)
  DEF_SAN(MouseOver)
  DEF_SAN(MouseOut)
  DEF_SAN(MouseMove)
  DEF_SAN(Hover)
  DEF_SAN(Unhover)
  DEF_SAN(LButtonDown)
  DEF_SAN(LButtonUp)
  DEF_SAN(LButtonClick)
  DEF_SAN(RButtonDown)
  DEF_SAN(RButtonUp)
  DEF_SAN(RButtonClick)
  DEF_SAN(MButtonDown)
  DEF_SAN(MButtonUp)
  DEF_SAN(MButtonClick)
  DEF_SAN(DropFile)
  DEF_SAN(DropText)
  DEF_SAN(DropURL)

  /* module only */ 
  DEF_SAN(Init)
  DEF_SAN(Quit)

  NULL
};

#define DECL_SAD(def) static size_t saDef##def(rcVarPointers *, TCHAR *, size_t, DWORD);
#define DEF_SAD(def) size_t saDef##def(rcVarPointers *var, TCHAR *dest, size_t maxchars, DWORD lParam)
DECL_SAD(WindowName)
DECL_SAD(DropItem)

#define USE_SAD(def) &saDef##def,
#define USE_SAD_NONE NULL,

rcVarLookupProc ScriptActionDefaults[ScriptActionModuleCount] =
{
  /* These MUST be in the same order as the */ 
  /* entries in the ScriptActionTypes enum  */ 

  /* module or window */ 
  USE_SAD_NONE /* Create */ 
  USE_SAD_NONE /* Destroy */ 
  USE_SAD_NONE /* MouseOver */ 
  USE_SAD_NONE /* MouseOut */ 
  USE_SAD_NONE /* MouseMove */ 
  USE_SAD_NONE /* Hover */ 
  USE_SAD_NONE /* Unhover */ 
  USE_SAD_NONE /* LButtonDown */ 
  USE_SAD_NONE /* LButtonUp */ 
  USE_SAD_NONE /* RButtonDown */ 
  USE_SAD_NONE /* RButtonUp */ 
  USE_SAD_NONE /* RButtonClick */ 
  USE_SAD_NONE /* MButtonDown */ 
  USE_SAD_NONE /* MButtonUp */ 
  USE_SAD_NONE /* MButtonClick */ 
  USE_SAD(DropItem) /* DropFile */ 
  USE_SAD(DropItem) /* DropText */ 
  USE_SAD(DropItem) /* DropURL */ 

  /* module only */ 
  USE_SAD_NONE /* Init */ 
  USE_SAD_NONE /* Quit */ 
};


DEF_SAD(WindowName)
{
  WndInfo *w = reinterpret_cast<WndInfo *>(lParam);
  if (!maxchars) return 0;
  ASSERT_PTR(dest, WRITEABLEN(dest,maxchars));
  ASSERT_PTR(w, READABLE(w));

  strncpy(dest, w->Name, maxchars-1);
  dest[maxchars] = 0;
  return strlen(dest);
}

DEF_SAD(DropItem)
{
  WndInfo *w = reinterpret_cast<WndInfo *>(lParam);
  ASSERT_PTR(dest, WRITEABLEN(dest,maxchars));
  ASSERT_PTR(w, READABLE(w));

  if (w->DropItem && *w->DropItem && maxchars)
  {
    strncpy(dest, w->DropItem, maxchars-1);
    dest[maxchars] = 0;
    return strlen(dest);
  }
  return 0;
}

bool MakeRelCoord(char rel,POINT *pt,WndInfo *w)
{
  RECT r;

  switch (rel)
  {
  case 'p':
  case 'P': /* rel to parent  */ 
    if (!w) return false;
    GetWindowRect(w->hParentWnd,&r);
    pt->x -= r.left;
    pt->y -= r.top;
    break;
  case 'm':
  case 'M': /* rel to monitor */ 
    if (!w) return false;
    LSGetMonitorScreenRect(w->Pos.Monitor,&r);
    pt->x -= r.left;
    pt->y -= r.top;
    break;
  case 'd':
  case 'D': /* rel to desktop */ 
    /* 0,0 is on top,leftmost monitor */ 
    LSGetMonitorScreenRect(LSMONITOR_ALL,&r);
    pt->x -= r.left;
    pt->y -= r.top;
    break;
  case 's':
  case 'S': /* rel to screen  */ 
    /* 0,0 is on primary monitor */ 
    break;
  default:
    return false;
  }
  return true;
}

size_t LookupVar(rcVarPointers *var, TCHAR *dest, size_t maxchars, DWORD lParam)
//size_t LookupVar(TCHAR *src, TCHAR *dest, size_t maxchars, DWORD lParam)
//int LookupVar(char *dest,int maxchars,const char *src,LPARAM lParam)
{
  TCHAR *src;
  LPARAM values[4];
  POINT pt;
  RECT r;
  DWORD d;
  bool bIntValue = false;
  bool left,right;
  int i;
  WndInfo *w = reinterpret_cast<WndInfo *>(lParam);

  ASSERT_PTR(var, READABLE(var));
  src = var->NameStart;
  ASSERT_PTR(src, READABLE(src));
  ASSERT_PTR(dest, WRITEABLEN(dest,maxchars));

  switch (*src++)
  {
	case '_': /* module info */ 
		switch (*src++) {
    case 'N':
		case 'n': lstrcpyn(dest,ModuleName,maxchars); break;
    case 'V':
		case 'v': lstrcpyn(dest,ModuleVersion,maxchars); break;
    case 'A':
		case 'a': lstrcpyn(dest,ModuleAuthor,maxchars); break;
    case 'B':
    case 'b': lstrcpyn(dest,ModuleBuild,maxchars); break;
    default: return 0;
		}
		break;

  case '#': /* special char */ 
    switch (*src)
    {
    case 'A':
    case 'a': *dest = '\''; break;
    case 'C':
    case 'c': *dest = '^'; break;
    case 'D':
    case 'd': *dest = '$'; break;
    case 'N':
    case 'n': *dest = '\n'; break;
    case 'Q':
    case 'q': *dest = '"'; break;
    default:  *dest = *src;
    }
    return 1;

  case 'K':
	case 'k': /* keys */ 
		switch (*src++) {
    case 'A':
		case 'a': /* alt              */ 
			left = IsKeyDown(VK_LMENU);
			right = IsKeyDown(VK_RMENU);
			break;
    case 'C':
		case 'c': /* ctrl             */ 
			left = IsKeyDown(VK_LCONTROL);
			right = IsKeyDown(VK_RCONTROL);
			break;
    case 'M':
		case 'm': /* context menu key */ 
			left = right = IsKeyDown(VK_APPS);
			break;
    case 'S':
		case 's': /* shift            */ 
			left = IsKeyDown(VK_LSHIFT);
			right = IsKeyDown(VK_RSHIFT);
			break;
    case 'W':
		case 'w': /* win key          */ 
			left = IsKeyDown(VK_LWIN);
			right = IsKeyDown(VK_RWIN);
			break;
    default: return 0;
		}
		i = 3;
		switch (*src++) {
		case 'l': i=1; break;
		case 'r': i=2; break;
		}
		if ((i&1)==0) left = false;
		if ((i&2)!=0) left |= right;
		/* result */ 
		if (left) lstrcpyn(dest,"true",maxchars);
    else lstrcpyn(dest,"false",maxchars);
		break;

  case 'M':
  case 'm': /* mouse */ 
    d = GetMessagePos();
    pt.x = (short)LOWORD(d);
    pt.y = (short)HIWORD(d);
    switch (*src++)
    {
    case 'X':
    case 'x':
    case 'Y':
    case 'y':
      if (!MakeRelCoord(*src,&pt,w))
        if (w) ScreenToClient(w->hWnd,&pt);
      bIntValue = true;
      values[0] = (src[-1]=='x'||src[-1]=='X')?pt.x:pt.y;
      break;
    default: return 0;
    }
    break;

  case 'I':
  case 'i': /* window */ 
    /* return nothing if no window */ 
    if (!w) return 0;
    GetWindowRect(w->hWnd,&r);
    switch (*src++)
    {
    case 'X':
    case 'x':
    case 'Y':
    case 'y':
      pt.x = r.left;
      pt.y = r.top;
      if (!MakeRelCoord(*src,&pt,w))
      {
        if (w->bOnTop || !WndInBox(w))
          /* rel to monitor */ 
          MakeRelCoord('m',&pt,w);
        else
          /* rel to parent */ 
          MakeRelCoord('p',&pt,w);
      }
      bIntValue = true;
      values[0] = (src[-1]=='x'||src[-1]=='X')?pt.x:pt.y;
      break;
    case 'S':
    case 's': /* size */ 
      switch (*src++)
      {
      case 'X':
      case 'x': /* width */ 
        values[0] = r.right-r.left;
        bIntValue = true;
        break;
      case 'Y':
      case 'y': /* height */ 
        values[0] = r.bottom-r.top;
        bIntValue = true;
        break;
      default: return 0;
      }
      break;
    case 'H':
    case 'h': /* handle */ 
      if (!w) return 0;
      values[0] = reinterpret_cast<LPARAM>(w->hWnd);
      bIntValue = true;
      break;
    case 'M':
    case 'm': /* monitor */ 
      if (!w) return 0;
      lstrcpyn(dest,LSGetMonitorName(w->Pos.Monitor),maxchars);
      break;
#ifndef SINGLE_WINDOW
    case 'N':
    case 'n':
      if (!w) return 0;
      lstrcpyn(dest,w->Name,maxchars);
      break;
    case 'G':
    case 'g':
      if (!w) return 0;
      values[0] = w->Group;
      bIntValue = true;
      break;
#endif
#ifndef NO_SKIN
    case 'B':
    case 'b':
      if (!w) return 0;
      values[0] = reinterpret_cast<LPARAM>(w->hbmBackBuffer);
      bIntValue = true;
      break;
    case 'K':
    case 'k':
      if (!w) return 0;
      values[0] = reinterpret_cast<LPARAM>(w->emSkin.Bitmap);
      bIntValue = true;
      break;
#endif
    default: return 0;
    }
    break;

  default: return 0;
  }

  if (bIntValue)
  {
    FormatMessage(
      FORMAT_MESSAGE_FROM_STRING|FORMAT_MESSAGE_ARGUMENT_ARRAY,
      "%1!d!",
      NULL,NULL,dest,maxchars,
      reinterpret_cast<char**>(values));
  }

  return lstrlen(dest);
}

static rcVarLookup VarLookup =
{ sizeof(rcVarLookup), LookupVar, NULL };

// execute an action by string, in a given window's context
void ExecActionString(HWND hCaller, WndInfo *w, const char *sAction, rcVarLookupProc defproc)
{
  char buffer[MAX_LINE_LENGTH];
  if (sAction)
  {
    ASSERT_PTR(sAction, READABLE(sAction));
    VarLookup.Def = defproc;
    TRACE_MSG(("ActionString: %1", sAction));
    rcVarReplace(const_cast<char *>(sAction),
      buffer, MAX_LINE_LENGTH,
      &VarLookup, reinterpret_cast<LPARAM>(w));
    TRACE_MSG(("Vars replaced: %1", buffer));
    LSExecute(hCaller,buffer,SW_SHOWNORMAL);
    //TRACE(0,"LSExecute returned.");
  }
}

// execute an action by index (checks range for window actions *only*)
void ExecAction(WndInfo *w, int iAction)
{
  const char *sAction = NULL;

  ASSERT(iAction<ScriptActionModuleCount);

#ifndef SINGLE_WINDOW
  if (w && iAction<ScriptActionWindowCount)
  {
    ASSERT_PTR(w, READABLE(w));
    if (w->Actions)
    {
      ASSERT_PTR(w->Actions, READABLEN(w->Actions,ScriptActionWindowCount));
      sAction = w->Actions[iAction];
    }
  }

  /* only exec global action if no specific one */ 
  if (!sAction)
#endif

  if (Module.Actions)
  {
    ASSERT_PTR(Module.Actions, READABLEN(Module.Actions,ScriptActionModuleCount));
    sAction = Module.Actions[iAction];
  }

  ExecActionString(w?w->hWnd:NULL, w, sAction, ScriptActionDefaults[iAction]);
}

// set or clear an action on a window or on the module
bool SetAction(WndInfo *w,int iAction,const char *Command)
{
  ScriptActionSet *pas;
  int nActions;

#ifndef SINGLE_WINDOW
  if (w)
  {
    /* set window action */ 
    pas = &w->Actions;
    nActions = ScriptActionWindowCount;
  }
  else
#endif
  {
    /* set module action */ 
    pas = &Module.Actions;
    nActions = ScriptActionModuleCount;
  }

  if (iAction>=nActions) return false;

  return SetAction(pas,nActions,iAction,Command);
}

// set or clear an action in a set (does not check range)
bool SetAction(ScriptActionSet *pas,int nActions,int iAction,const char *Command)
{
  REQUIRE(iAction<nActions);

  /* clear any existing action */ 
  if (*pas)
  {
    ASSERT_PTR(*pas, WRITEABLEN(*pas,nActions));
    if ((*pas)[iAction])
    {
      ASSERT_PTR(*pas[iAction], WRITEABLE((*pas)[iAction]));
      delete[] (*pas)[iAction];
      (*pas)[iAction] = NULL; /* used as flag */ 
    }
  }

  if (Command && *Command)
  {
    if (*pas==NULL)
    {
      *pas = new ScriptAction[nActions];
      if (!*pas) return false;
      ASSERT_PTR(*pas, WRITEABLEN(*pas,nActions));
      ZeroMemory(*pas,nActions*sizeof(ScriptAction));
    }

    (*pas)[iAction] = new char[lstrlen(Command)+1];
    if (!*pas) return false;
    ASSERT_PTR((*pas)[iAction], WRITEABLEN((*pas)[iAction],(lstrlen(Command)+1)));
    lstrcpy((*pas)[iAction],Command);
  }

  return true;
}

#endif /* NO_SCRIPT_ACTIONS */ 