#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <ls-b247\lsapi\lsapi.h>
#include "lsb_contract.h"
#include "_lsb_dbghelp.h"

#ifndef NDEBUG

// ASSERTs, etc. are not allowed in this file! 
// That could cause infinite recursion.        


LSBASE_NAMESPACE::Logger::Logger(const char *sPrefix, const char *sFileName)
 : nInTrace_(0), bEnabled_(true)
{
  char sVarName[256];
  /* $<ModulePrefix>LogFile$              */ 
  lstrcpy(sVarName, sPrefix);
  lstrcat(sVarName, "LogFile");
  if (GetRCLine(sVarName, sLogPath_, MAX_LINE_LENGTH, NULL))
    ;
  /* $LSLogPath$<ModuleName>.log        */ 
  else if (GetRCLine("LSLogPath", sLogPath_, MAX_LINE_LENGTH, NULL))
  {
    lstrcat(sLogPath_, sFileName);
    lstrcat(sLogPath_, ".log");
  }
  /* $LiteStepDir$logs\<ModuleName>.log */ 
  else if (GetRCLine("LiteStepDir", sLogPath_, MAX_LINE_LENGTH, NULL))
  {
    lstrcat(sLogPath_, "logs\\");
    CreateDirectory(sLogPath_, NULL);
    lstrcat(sLogPath_, sFileName);
    lstrcat(sLogPath_, ".log");
  }
}

bool LSBASE_NAMESPACE::Logger::RawWrite_(const char * sLine)
{
  /* open the log file */ 
  HANDLE hFile =
    CreateFile(sLogPath_, FILE_APPEND_DATA, FILE_SHARE_READ, NULL,
      OPEN_ALWAYS, FILE_ATTRIBUTE_ARCHIVE, NULL);

  /* turn off logging if we couldn't open the file */ 
  if (!hFile)
  {
    bEnabled_ = false;
    return false;
  }

  /* write to the end of the file and finish */ 
  SetFilePointer(hFile,0,0,FILE_END);

  DWORD bWritten;

  /* actual message */ 
  WriteFile(hFile, sLine, lstrlen(sLine),
    &bWritten, NULL);
  WriteFile(hFile, "\r\n", 2*sizeof(char),
    &bWritten, NULL);

  CloseHandle(hFile);

  return true;
}

bool LSBASE_NAMESPACE::Logger::WriteTimeStamp_(void)
{
  /* date/time stamp */ 
  SYSTEMTIME SystemTime;
  GetLocalTime(&SystemTime);

  LPSTR values[] =
  {
    (LPSTR)SystemTime.wYear, (LPSTR)SystemTime.wMonth,
    (LPSTR)SystemTime.wDay, (LPSTR)SystemTime.wHour,
    (LPSTR)SystemTime.wMinute, (LPSTR)SystemTime.wSecond,
    (LPSTR)SystemTime.wMilliseconds
  };

  FormatMessage(
    FORMAT_MESSAGE_FROM_STRING|FORMAT_MESSAGE_ARGUMENT_ARRAY,
    "-- %1!04d!-%2!02d!-%3!02d! %4!02d!:%5!02d!:%6!02d!.%7!03d!",
    NULL,NULL,sLineBuffer_,MAX_LINE_LENGTH,values);

  return RawWrite_(sLineBuffer_);
}

bool LSBASE_NAMESPACE::Logger::WriteStack_(unsigned nDepth)
{
#ifdef USE_DBGHELP
  /* short stack trace */ 
  StackTrace st(*reinterpret_cast<CONTEXT*>(LSBASE_NAMESPACE::g_GrabbedContext));
  if (st.Init())
  {
    while (nDepth-- && st.Walk())
    {
      st.GetCurrentPosition(sLineBuffer_, MAX_LINE_LENGTH);
      if (!RawWrite_(sLineBuffer_)) return false;
    }
  }
#endif
  return true;
}

void LSBASE_NAMESPACE::Logger::operator()(const char * sFormat, ...)
{
  bool bOk = Begin();

  try
  {
    /* format the error message */ 
    va_list pArgs;
    va_start(pArgs, sFormat);

    FormatMessage(FORMAT_MESSAGE_FROM_STRING, sFormat,
      NULL, NULL, sLineBuffer_, MAX_LINE_LENGTH, &pArgs);

    va_end(pArgs);

    if (!bOk)
    {
      LPSTR values[2] = { sLogPath_, sLineBuffer_ };
      char sBuffer[MAX_LINE_LENGTH];

      FormatMessage(
        FORMAT_MESSAGE_FROM_STRING|FORMAT_MESSAGE_ARGUMENT_ARRAY,
        "Unable to open log file:\n%1\nLogging will be disabled.\n\nMessage was:\n%2",
        NULL,NULL,sBuffer,MAX_LINE_LENGTH,values);
      MessageBox(NULL, sBuffer, "LiteStep Module", MB_OK|MB_ICONERROR);

      return;
    }

    RawWrite_(sLineBuffer_);
  }
  catch(...)
  {
  }

  End();
}

bool LSBASE_NAMESPACE::Logger::Begin(void)
{
  bool bRet = false;
  try
  {
    if (!nInTrace_++)
      bRet = WriteTimeStamp_() && WriteStack_(3);
    else
      bRet = true;
  }
  catch(...)
  {
  }
  return bRet;
}

bool LSBASE_NAMESPACE::Logger::End(void)
{
  bool bRet = false;
  try
  {
    if (!--nInTrace_)
    {
      bRet = RawWrite_("");
    }
    else
      bRet = true;
  }
  catch(...)
  {
  }
  return bRet;
}

bool LSBASE_NAMESPACE::Logger::ForceBlock(void)
{
  if (nInTrace_)
  {
    return RawWrite_("")
      && WriteTimeStamp_()
      && WriteStack_(3);
  }
  else
    return true;
}

////
///
//

LSBASE_NAMESPACE::CheckFailed::CheckFailed
  (Logger &Log, const char * sType, const char * sCondition)
{
  Log.ForceBlock();
  Log("CheckFailed: %1(%2)", sType, sCondition);
}

LSBASE_NAMESPACE::AssertPointerFailed::AssertPointerFailed(
      Logger &Log, const char * sType, const char * sCondition,
      const void *pPointer, const char * sName)
  : sType_(sType), sCondition_(sCondition), pPointer_(pPointer), sName_(sName)
{
  Log.ForceBlock();
  Log.Begin();
  Log("CheckFailed: %1(%2)", sType, sCondition);
  Log("  %1 = %2!08X!", sName, pPointer);

  HANDLE hProcess = GetCurrentProcess();
  SymInitialize(hProcess, NULL, TRUE);
  union
  {
    IMAGEHLP_SYMBOL64 sym;
    DWORD symbol_buffer[1024];
  };
  DWORD64 dOffset, dPointer = (DWORD64)pPointer;

  ZeroMemory(&sym, sizeof(symbol_buffer));
  sym.SizeOfStruct = sizeof(symbol_buffer);
  sym.MaxNameLength = sizeof(symbol_buffer)-sizeof(sym);
  dOffset = 0;

  if (SymGetSymFromAddr64(
        hProcess,
        dPointer,
        &dOffset,
        &sym))
  {
    Log("  %1 = %2!08X! + %3!d! bytes", sym.Name, sym.Address, sym.Size);
  }
  else
  {
    sym.Size = 1;
    Log("  No symbol data.");
  }

  SymCleanup(hProcess);

  if (!IsBadStringPtr(reinterpret_cast<LPCTSTR>(pPointer),sym.Size))
  {
    Log("  STRING \"%1\"", pPointer);
  }
  if (READABLEB(pPointer,sym.Size))
  {
    if (sym.Size>4)
    {
      const DWORD *p = reinterpret_cast<const DWORD *>(pPointer);
      if (sym.Size>16)
        Log("  -> %1!08X! %2!08X! %3!08X! %4!08X!", p[0], p[1], p[2], p[3]);
      else if (sym.Size>12)
        Log("  -> %1!08X! %2!08X! %3!08X!", p[0], p[1], p[2]);
      else if (sym.Size>8)
        Log("  -> %1!08X! %2!08X!", p[0], p[1]);
      else
        Log("  -> %1!08X!", p[0]);
    }
    Log("  READABLE");
  }
  if (WRITEABLEB(const_cast<void*>(pPointer),sym.Size))
    Log("  WRITEABLE");
  Log("/");
  Log.End();
}

#endif
