/*
  RabidVWM LiteStep Module - version 1.0
	IDropTarget Implementation for Drag and Drop

  Copyright (C) 2002 Joshua Seagoe
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "module.h"
#include "droptarget.h"
#include <Shlobj.h>
#include <ShellAPI.h>

#define USE_PIDL


UINT SkinBoxDropTarget::cf_types[] =
{
	0, /* CFSTR_SHELLIDLIST */ 
	CF_HDROP,
	0, /* CFSTR_INETURL			*/ 
	0, /* CFSTR_SHELLURL		*/ 
	CF_TEXT,
	CF_UNICODETEXT,
	0
};

/*

	CF_HDROP:
		works for regular files
		does not work for folders
		never sent for "special" things

	CFSTR_SHELLIDLIST:
		works for regular files
		works for folders
		works for "special" as for corresponding regular type

*/

/* It's like a switch statement, only greasier! */ 
SkinBoxDropTarget::DropFuncPtr SkinBoxDropTarget::cf_actions[] =
{
	&SkinBoxDropTarget::Drop_SHELLIDLIST,
	&SkinBoxDropTarget::Drop_HDROP,
	&SkinBoxDropTarget::Drop_INETURL,
	&SkinBoxDropTarget::Drop_SHELLURL,
	&SkinBoxDropTarget::Drop_TEXT,
	&SkinBoxDropTarget::Drop_UNICODETEXT,
	0
};

void SkinBoxDropTarget::RegisterFormats(void)
{
	// Why did Microsoft think it would be a good idea to force people to do this?
	cf_types[0] = RegisterClipboardFormat(CFSTR_SHELLIDLIST);
	cf_types[2] = RegisterClipboardFormat(CFSTR_INETURL);
	cf_types[3] = RegisterClipboardFormat(CFSTR_SHELLURL);
}


HRESULT STDMETHODCALLTYPE SkinBoxDropTarget::QueryInterface( 
	/* [in] */ REFIID riid,
	/* [iid_is][out] */ void __RPC_FAR *__RPC_FAR *ppvObject)
{
	return E_NOINTERFACE;
}

ULONG STDMETHODCALLTYPE SkinBoxDropTarget::AddRef( void)
{
	return 1;
}

ULONG STDMETHODCALLTYPE SkinBoxDropTarget::Release( void)
{
	return 1;
}

bool SkinBoxDropTarget::CheckDropType(FORMATETC &etc, IDataObject *pDataObj)
{
  if (!pWnd->Actions) return false;

	ZeroMemory(&etc,sizeof(FORMATETC));

	/* We want all of the content, in a global handle. */ 
	etc.tymed = TYMED_HGLOBAL;
	etc.dwAspect = DVASPECT_CONTENT;
	etc.lindex = -1;

	cf_index = 0;
  // can accept a file
  if (pWnd->Actions[OnDropFile] && *pWnd->Actions[OnDropFile])
  {
    cf_index = 0;
		etc.cfFormat = cf_types[cf_index];
		if (pDataObj->QueryGetData(&etc)==S_OK)
      return true;
    cf_index = 1;
		etc.cfFormat = cf_types[cf_index];
		if (pDataObj->QueryGetData(&etc)==S_OK)
      return true;
  }
  // can accept a url
  if (pWnd->Actions[OnDropURL] && *pWnd->Actions[OnDropURL])
  {
    cf_index = 2;
		etc.cfFormat = cf_types[cf_index];
		if (pDataObj->QueryGetData(&etc)==S_OK)
      return true;
    cf_index = 3;
		etc.cfFormat = cf_types[cf_index];
		if (pDataObj->QueryGetData(&etc)==S_OK)
      return true;
  }
  // can accept text
  if (pWnd->Actions[OnDropText] && *pWnd->Actions[OnDropText])
  {
    cf_index = 4;
		etc.cfFormat = cf_types[cf_index];
		if (pDataObj->QueryGetData(&etc)==S_OK)
      return true;
    cf_index = 5;
		etc.cfFormat = cf_types[cf_index];
		if (pDataObj->QueryGetData(&etc)==S_OK)
      return true;
  }
  return false;
}

HRESULT STDMETHODCALLTYPE SkinBoxDropTarget::DragEnter( 
	/* [unique][in] */ IDataObject *pDataObj,
	/* [in] */ DWORD grfKeyState,
	/* [in] */ POINTL pt,
	/* [out][in] */ DWORD *pdwEffect)
{
  FORMATETC etc;
	TRACE("DragEnter");

  SendMessage(pWnd->hWnd, WM_MOUSEMOVE, (WPARAM)grfKeyState, MAKELONG(pt.x,pt.y));

	*pdwEffect = DROPEFFECT_NONE;
  if (CheckDropType(etc,pDataObj))
    *pdwEffect = DROPEFFECT_COPY;
  else
    cf_index = -1;

  return S_OK;
}

HRESULT STDMETHODCALLTYPE SkinBoxDropTarget::DragOver( 
	/* [in] */ DWORD grfKeyState,
	/* [in] */ POINTL pt,
	/* [out][in] */ DWORD *pdwEffect)
{
	// forward it to the window to get hovering
	// mouse messages don't get sent to the main window during drag&drop
  SendMessage(pWnd->hWnd, WM_MOUSEMOVE, (WPARAM)grfKeyState, MAKELONG(pt.x,pt.y));

  if (cf_index == -1)
    *pdwEffect = DROPEFFECT_NONE;
  else
    *pdwEffect = DROPEFFECT_COPY;

	return S_OK;
}

HRESULT STDMETHODCALLTYPE SkinBoxDropTarget::DragLeave( void)
{
	TRACE("DragLeave");

  SendMessage(pWnd->hWnd, WM_MOUSEMOVE, 0, 0);

	return S_OK;
}

HRESULT STDMETHODCALLTYPE SkinBoxDropTarget::Drop( 
	/* [unique][in] */ IDataObject *pDataObj,
	/* [in] */ DWORD grfKeyState,
	/* [in] */ POINTL pt,
	/* [out][in] */ DWORD *pdwEffect)
{
	FORMATETC etc;
	STGMEDIUM stg;
	ZeroMemory(&etc,sizeof(FORMATETC));

  SendMessage(pWnd->hWnd, WM_MOUSEMOVE, (WPARAM)grfKeyState, MAKELONG(pt.x,pt.y));

	/* We want all of the content, in a global handle. */ 
	etc.tymed = TYMED_HGLOBAL;
	etc.dwAspect = DVASPECT_CONTENT;
	etc.lindex = -1;

	etc.cfFormat = cf_index;
	if (pDataObj->QueryGetData(&etc)!=S_OK)
	{
		// return if we can't process any data types provided
    if (!CheckDropType(etc, pDataObj))
      return S_OK;
	}

	/* Grab the stuff */ 
	if (pDataObj->GetData(&etc,&stg)==S_OK)
	{
		/* Use it, locked so we get a memory reference. */ 
		(this->*cf_actions[cf_index])(GlobalLock(stg.hGlobal));
		GlobalUnlock(stg.hGlobal);

		/* Free the medium. */ 
		if (stg.pUnkForRelease==NULL)
			GlobalFree(stg.hGlobal);
		else
			stg.pUnkForRelease->Release();
	}

  cf_index = -1;

	return S_OK;
}

#define HIDA_GetPIDLParent(pida) (LPCITEMIDLIST)(((LPBYTE)pida)+(pida)->aoffset[0])
#define HIDA_GetPIDLItem(pida, i) (LPCITEMIDLIST)(((LPBYTE)pida)+(pida)->aoffset[i+1])

#ifdef USE_PIDL

void DumpPIDL(LPCITEMIDLIST pidl)
{
#if !defined(NDEBUG) && defined(TRACE_FILE)
	char buffer[MAX_PATH];
	int j;
#endif
#if !defined(NDEBUG) && defined(TRACE_FILE)
		sprintf(buffer,"PIDL:%08X",pidl);
		TRACE(TRACE_DUMP,buffer);

		j = ((SHITEMID *)(pidl))->cb;
		buffer[j+j] = 0;
		while (j--)
		{
			/* convert to hex (scary lookin, innit?) */ 
			buffer[j+j+1] = (reinterpret_cast<const SHITEMID *>(pidl)->abID[j]&0x0f)["0123456789ABCDEF"];
			buffer[j+j+0] = (reinterpret_cast<const SHITEMID *>(pidl)->abID[j] >> 4)["0123456789ABCDEF"];
		}
		TRACE(TRACE_DUMP,buffer);

		j = (reinterpret_cast<const SHITEMID *>(pidl))->cb;
		buffer[j] = 0;
		while (j--)
		{
			buffer[j] = (reinterpret_cast<const SHITEMID *>(pidl)->abID[j]);
			if (buffer[j]<32 || buffer[j]>126) buffer[j] = '-';
		}
		TRACE(TRACE_DUMP,buffer);

		SHGetPathFromIDList(reinterpret_cast<LPCITEMIDLIST>(pidl),buffer);
		TRACE(TRACE_DUMP,buffer);
#endif
}

UINT GetSizePIDL(LPCITEMIDLIST pidl)
{
	UINT i = (pidl->mkid.cb);

	if (i==0) return sizeof(pidl->mkid.cb);

	TRACE("** sizing");
	DumpPIDL(pidl);

	// walk the id list chain
	return i + GetSizePIDL(
		reinterpret_cast<LPCITEMIDLIST>(
		reinterpret_cast<const BYTE *>(pidl) + i
		));
}

LPITEMIDLIST AppendPIDL(LPMALLOC pMalloc,LPCITEMIDLIST pidlBase, LPCITEMIDLIST pidlAdd)
{
    if(pidlBase == NULL)
        return NULL;
    if(pidlAdd == NULL)
        return const_cast<LPITEMIDLIST>(pidlBase);
    
    LPITEMIDLIST pidlNew;

/* Microsoft has an example for this, but it only works
   if the id lists contain only one element each.
	 (either that, or they do exactly the same thing as
	 this, but don't tell you how to get GetSizePIDL)
*/

    UINT cb1 = GetSizePIDL(pidlBase);
    UINT cb2 = GetSizePIDL(pidlAdd);

    pidlNew = (LPITEMIDLIST)pMalloc->Alloc(cb1 + cb2);
    if (pidlNew)
    {
        CopyMemory(((LPSTR)pidlNew), pidlBase, cb1);
        CopyMemory(((LPSTR)pidlNew) + cb1 - sizeof(pidlBase->mkid.cb), pidlAdd, cb2);

				TRACE("-- result");
				DumpPIDL(pidlNew);
    }
    return pidlNew;
}

#endif /* USE_PIDL */ 


void SkinBoxDropTarget::Drop_SHELLIDLIST(void *files)
{
  ITEMIDLIST *lpIDList;
	CIDA *pCIDA = reinterpret_cast<CIDA *>(files);
	UINT i = 0;
  char DropItem[MAX_PATH];

	TRACE("Drop_SHELLIDLIST");

	LPMALLOC pMalloc;

	SHGetMalloc(&pMalloc);

  pWnd->DropItem = DropItem;
  pWnd->nDropLeft = pCIDA->cidl;
	while (i<pCIDA->cidl)
	{
		// stupid shell gives me a folder and a relative pidl
		// where i really need an absolute pidl
		lpIDList = AppendPIDL(pMalloc,
			HIDA_GetPIDLParent(pCIDA),HIDA_GetPIDLItem(pCIDA,i));

    SHGetPathFromIDList(lpIDList, DropItem);
    --pWnd->nDropLeft;
    ExecAction(pWnd, OnDropFile);

		pMalloc->Free(lpIDList);

		++i;
	}
  pWnd->DropItem = NULL;

	pMalloc->Release();
}

int CountHDROPItems(const wchar_t *names)
{
  int count = 0;
  while (*names!=0)
  {
    ++count;
		while (*++names!=0) ;
		++names;
  }
  return count;
}

void SkinBoxDropTarget::Drop_HDROP(void *files)
{
	DROPFILES *df = reinterpret_cast<DROPFILES *>(files);
	wchar_t *names;
	char name[MAX_PATH_LENGTH];

	TRACE("Drop_HDROP");

	// damn unicode
	names = reinterpret_cast<wchar_t *>(reinterpret_cast<BYTE *>(files)+df->pFiles);
  pWnd->DropItem = name;
  pWnd->nDropLeft = CountHDROPItems(names);
	while (*names!=0)
	{
		WideCharToMultiByte(CP_ACP,0,names,-1,name,MAX_PATH_LENGTH,NULL,NULL);

		TRACE(name);
    --pWnd->nDropLeft;
    ExecAction(pWnd, OnDropFile);

		while (*++names!=0) ;
		++names;
	}
  pWnd->DropItem = NULL;
}

void SkinBoxDropTarget::Drop_INETURL(void *text)
{
	char *p = reinterpret_cast<char *>(text);
	char name[MAX_PATH_LENGTH]={0};
	TRACE("Drop_INETURL");

	if (p[1]==0)
	{
		/* assume urls are never 1 byte long */ 
		/* so this one's in unicode          */ 
		WideCharToMultiByte(CP_ACP,0,reinterpret_cast<wchar_t *>(text),-1,name,MAX_PATH_LENGTH,NULL,NULL);
		p = name;
	}
	else
	{
		p = reinterpret_cast<char *>(text);
	}

  pWnd->DropItem = p;
  pWnd->nDropLeft = 0;
  ExecAction(pWnd, OnDropURL);
	pWnd->DropItem = NULL;
}

void SkinBoxDropTarget::Drop_SHELLURL(void *text)
{
	TRACE("Drop_SHELLURL");
	Drop_INETURL(text);
}

int CountTEXTItems(const char *text, char nl)
{
  int nItems = 1;
  while (*text!=0)
  {
    if (*text == nl) ++nItems;
    ++text;
    while (*text=='\n' || *text=='\r')
      ++text;
  }
  return nItems;
}

void SkinBoxDropTarget::Drop_TEXT(void *text)
{
	char *p = reinterpret_cast<char *>(text);
	char *eol, *eos;
	char nl = '\r';
  char item[MAX_LINE_LENGTH];

	TRACE("Drop_TEXT");

  // find the right end of line char
  // should be \r\n according to msdn, but
  // since when does everything play nice?
	eol = strchr(p,'\r');
  eos = strchr(p,'\n');
  if (!eol || eos<eol) { nl='\n'; eol=eos; }
  pWnd->nDropLeft = CountTEXTItems(p, nl);

  eos = p;
  while (*eos) ++eos;
	
	// parse it line-for-line and execute it
  pWnd->DropItem = item;
	while (*eol)
	{
    ASSERT(eol<eos);
    ASSERT(eol>=text);
/*
!skinboxaction one droptext !alert "Text: <{\\}>"
!skinboxaction one dropfile !alert "File: <{\\}>"
!skinboxaction one dropurl !alert "URL: <{\\}>"

!skinboxaction one droptext !commandsettext Text: <{}>
!skinboxaction one dropfile !commandsettext File: <{}>
!skinboxaction one dropurl !commandsettext URL: <{}>
*/
		*eol = 0;
    //pWnd->DropItem = p;
    strcpy(item,p);
    --pWnd->nDropLeft;
    ExecAction(pWnd, OnDropText);
    ASSERT(eol<=eos);
    ASSERT(eol>text);

    p = eol;
    do ++p; while (*p=='\r' || *p=='\n');
    ASSERT(p<=eos);
    ASSERT(p>text);

    for (eol=p; *eol && *eol!=nl; ++eol) ;
    ASSERT(eol<=eos);
    ASSERT(eol>text);
	}
  ASSERT(strchr(p,nl)==NULL);
  pWnd->DropItem = p;
  --pWnd->nDropLeft;
  if (*p)
    ExecAction(pWnd, OnDropText);
  pWnd->DropItem = NULL;
}

// convert unicode text to mbcs and use Drop_TEXT
void SkinBoxDropTarget::Drop_UNICODETEXT(void *text)
{
	char buffer[MAX_LINE_LENGTH];

	TRACE("Drop_UNICODETEXT");

	WideCharToMultiByte(CP_ACP,0,reinterpret_cast<wchar_t *>(text),-1,buffer,
		MAX_LINE_LENGTH,NULL,NULL);

	Drop_TEXT(buffer);
}

