
#include <dbghelp.h>

/* note: this is the compatibility version of this file,
   to be used with modules not conforming to the new
   skinbox structure.
 */
#ifndef LSBASE_NAMESPACE
#define LSBASE_NAMESPACE lsbase
#endif

namespace LSBASE_NAMESPACE
{

#ifndef NDEBUG
  class Logger
  {
  protected:
    unsigned nInTrace_;
    char sLogPath_[MAX_LINE_LENGTH];
    char sLineBuffer_[MAX_LINE_LENGTH];
    bool bEnabled_;

    bool RawWrite_(const char *sLine);
    bool WriteTimeStamp_(void);
    bool WriteStack_(unsigned nDepth);

  public:
    Logger(const char *sPrefix, const char *sFileName);
    bool Begin(void);
    bool End(void);
    bool ForceBlock(void);
    void operator()(const char *sFormat, ...);
  };

  extern Logger g_ModuleLogger; /* to be defined by module */ 
#define DEFINE_MODULE_LOGGER(PREFIX, LOGFILENAME) LSBASE_NAMESPACE::Logger LSBASE_NAMESPACE::g_ModuleLogger(PREFIX, LOGFILENAME)
  extern DWORD g_GrabbedContext[sizeof(CONTEXT)/sizeof(DWORD)];

  typedef int offset_t;
#define OFFSET_OF(BASE, MEMBER) ((LSBASE_NAMESPACE::offset_t)&(((reinterpret_cast<BASE *>(0))->MEMBER)))
#define APPLY_OFFSET(OBJECT, TYPE, OFFSET) (*reinterpret_cast<TYPE *>((reinterpret_cast<BYTE *>(&OBJECT))+OFFSET))
  inline bool __declspec(naked) GrabContext(void)
  {
    /* "You cannot get a valid context for a running thread. Use */ 
    /* the SuspendThread function to suspend the thread before   */ 
    /* calling GetThreadContext."  (MSDN)                        */ 

    /* ...so we'll have to fake it.                              */ 
  
    enum {
      ieax = OFFSET_OF(CONTEXT,Eax),
      iebx = OFFSET_OF(CONTEXT,Ebx),
      iecx = OFFSET_OF(CONTEXT,Ecx),
      iedx = OFFSET_OF(CONTEXT,Edx),
      iesi = OFFSET_OF(CONTEXT,Esi),
      iedi = OFFSET_OF(CONTEXT,Edi),
      ics  = OFFSET_OF(CONTEXT,SegCs),
      ids  = OFFSET_OF(CONTEXT,SegDs),
      ies  = OFFSET_OF(CONTEXT,SegEs),
      ifs  = OFFSET_OF(CONTEXT,SegFs),
      igs  = OFFSET_OF(CONTEXT,SegGs),
      iss  = OFFSET_OF(CONTEXT,SegSs),
      iebp = OFFSET_OF(CONTEXT,Ebp),
      ieip = OFFSET_OF(CONTEXT,Eip),
      iesp = OFFSET_OF(CONTEXT,Esp),
      ieflags = OFFSET_OF(CONTEXT,EFlags),
      icxflags = OFFSET_OF(CONTEXT,ContextFlags)
    };
    __asm {
      pushfd
      push eax
      pushfd
      pop eax
      mov g_GrabbedContext[ieflags], eax
      pop eax
      mov g_GrabbedContext[ieax], eax
      xor eax,eax
      mov ax,gs
      mov g_GrabbedContext[igs], eax
      mov ax,fs
      mov g_GrabbedContext[ifs], eax
      mov ax,es
      mov g_GrabbedContext[ies], eax
      mov ax,ds
      mov g_GrabbedContext[iedi], edi
      mov g_GrabbedContext[iesi], esi
      mov g_GrabbedContext[iebx], ebx
      mov g_GrabbedContext[iedx], edx
      mov g_GrabbedContext[ids], eax
      mov g_GrabbedContext[iecx], ecx
      mov g_GrabbedContext[iebp], ebp
      mov eax, [esp+4]
      mov g_GrabbedContext[ieip], eax
      mov ax,cs
      mov g_GrabbedContext[ics], eax
      lea eax,[esp+8]
      mov g_GrabbedContext[iesp], eax
      mov ax,ss
      mov g_GrabbedContext[iss], eax

      mov eax, g_GrabbedContext[ieax]
      popfd
    }
    g_GrabbedContext[icxflags] = CONTEXT_SEGMENTS|CONTEXT_INTEGER|CONTEXT_CONTROL;
    __asm { ret }
  }
#endif /* NDEBUG */ 

  #ifndef NDEBUG
  class CheckFailed
  {
  public:
    CheckFailed(Logger &Log, const char *sType, const char * sCondition);
  };
  #define CHECK_FAILED(TYPE,CONDITION) LSBASE_NAMESPACE::CheckFailed(LSBASE_NAMESPACE::g_ModuleLogger,TYPE,CONDITION)
  #define CHECK_TEST(TYPE,TEST,CONDITION) if (!(TEST)) { LSBASE_NAMESPACE::GrabContext(); CHECK_FAILED(TYPE,CONDITION); } else

  class AssertFailed
  {
    const char * sType_, *sCondition_;
  public:
    AssertFailed(Logger &Log, const char * sType, const char * sCondition)
      : sType_(sType), sCondition_(sCondition)
    { CheckFailed(Log,sType,sCondition); }
  };
  #define ASSERT_FAILED(TYPE,CONDITION) LSBASE_NAMESPACE::AssertFailed(LSBASE_NAMESPACE::g_ModuleLogger,TYPE,CONDITION)
  #define ASSERT_THROW(TYPE,CONDITION) { LSBASE_NAMESPACE::GrabContext(); throw ASSERT_FAILED(TYPE,CONDITION); }
  #define ASSERT_TEST(TYPE,TEST,CONDITION) if (!(TEST)) ASSERT_THROW(TYPE,CONDITION) else
  #define ASSUME_TEST(TYPE,TEST,CONDITION) ASSERT_TEST(TEST,TYPE,CONDITION)

  class AssertPointerFailed
  {
    const char * sType_, *sCondition_, *sName_;
    const void *pPointer_;
  public:
    AssertPointerFailed(Logger &Log, const char * sType, const char * sCondition,
      const void *pPointer, const char * sName);
  };
  #define ASSERT_POINTER_FAILED(TYPE,CONDITION,POINTER,NAME) LSBASE_NAMESPACE::AssertPointerFailed(LSBASE_NAMESPACE::g_ModuleLogger,TYPE,CONDITION,POINTER,NAME)
  #define ASSERT_POINTER_THROW(TYPE,CONDITION,POINTER,NAME) { LSBASE_NAMESPACE::GrabContext(); throw ASSERT_POINTER_FAILED(TYPE,CONDITION,POINTER,NAME); }
  #define ASSERT_POINTER_TEST(TYPE,TEST,CONDITION,POINTER,NAME) if (!(TEST)) ASSERT_POINTER_THROW(TYPE,CONDITION,POINTER,NAME) else

  #define IFDEBUG(STATEMENTS) if (true) { STATEMENTS; } else
  #define DCATCHALL(STATEMENTS) if (true) try{STATEMENTS;}catch(...){} else
  #else /* NDEBUG */ 

  #define CHECK_TEST(TYPE,TEST,CONDITION) (void)0
  #define ASSERT_TEST(TYPE,TEST,CONDITION) (void)0
  #define ASSUME_TEST(TYPE,TEST,CONDITION) __assume(TEST)
  #define ASSERT_POINTER_TEST(TYPE,TEST,CONDITION,POINTER,NAME) (void)0
  #define IFDEBUG(STATEMENTS) (void)0
  #define DCATCHALL(STATEMENTS) if (true) { STATEMENTS; } else
  #endif /* NDEBUG */ 

  #ifndef NDEBUG
  class TraceBlock
  {
  public:
    TraceBlock(bool=false) { LSBASE_NAMESPACE::g_ModuleLogger.Begin(); }
    ~TraceBlock()          { LSBASE_NAMESPACE::g_ModuleLogger.End();   }
    operator bool()        { return false;              }
  };
  #define TRACE_BLOCK(NAME) LSBASE_NAMESPACE::TraceBlock NAME(LSBASE_NAMESPACE::GrabContext())
  #else
  #define TRACE_BLOCK(NAME) (void)0
  #endif
}

/* REQUIRE: preconditions for a function                         */ 
/* DO NOT use this in a destructor                               */ 
#define REQUIRE(CONDITION) ASSERT_TEST("REQUIRE",CONDITION,#CONDITION)
/* ENSURE: postconditions for a function                         */ 
/* You SHOULD use this in the constructor, but SHOULD NOT use it */ 
/* in the destructor                                             */ 
#define ENSURE(CONDITION) ASSERT_TEST("ENSURE",CONDITION,#CONDITION)
/* ASSERT: General purpose check.                                */ 
/* DO NOT use this in a destructor                               */ 
#define ASSERT(CONDITION) ASSERT_TEST("ASSERT",CONDITION,#CONDITION)
/* ASSUME: General purpose check with suggestion for optimizer   */ 
/* DO NOT use this in a destructor                               */ 
#define ASSUME(CONDITION) ASSUME_TEST("ASSUME",CONDITION,#CONDITION)
/* CHECK: General purpose check, do not interrupt if false       */ 
/* You MAY use this in a destructor                              */ 
#define CHECK(CONDITION) CHECK_TEST("CHECK",CONDITION,#CONDITION)

/* TRACE: Write a string to the log.                             */ 
#define TRACE(MESSAGE) TRACE_MSG(("%1",MESSAGE))

/* TRACE_MSG can be used to write a formatted string, but you    */ 
/* have to use two sets of parenthesis around the agument list.  */ 
/*   TRACE_MSG(("%1 = %2!d!", "x", 12));                         */ 
/* Note that this uses FormatMessage codes, not printf.          */ 
#define TRACE_MSG(PARAMS) IFDEBUG(LSBASE_NAMESPACE::GrabContext(); LSBASE_NAMESPACE::g_ModuleLogger PARAMS)

/* DCATCHALL: Use this if you need to call something from a      */ 
/* destructor that might throw an exception in debug builds.     */ 
/* ASSERT, etc. are safe from within functions called from       */ 
/* inside a DCATCHALL.                                           */ 

template<typename t>inline bool READABLEB(const t*p,unsigned b){return!IsBadReadPtr(p,b);}
template<typename t>inline bool READABLEN(const t*p,unsigned n){return READABLEB(p,n*sizeof(t));}
template<typename t>inline bool READABLE(const t*p){return READABLEN(p,1);}
template<typename t>inline bool READABLER(const t*s,const t*e){return(e>s)&&READABLEN(s,e-s);}
inline bool READABLER(const void *s,const void *e){return READABLER(reinterpret_cast<const BYTE *>(s),reinterpret_cast<const BYTE *>(e));}

template<typename t>inline bool WRITEABLEB(t*p,unsigned b){return!IsBadWritePtr(p,b);}
template<typename t>inline bool WRITEABLEN(t*p,unsigned n){return WRITEABLEB(p,n*sizeof(t));}
template<typename t>inline bool WRITEABLE(t*p){return WRITEABLEN(p,1);}
template<typename t>inline bool WRITEABLER(t*s,const t*e){return(e>s)&&WRITEABLEN(s,e-s);}
inline bool WRITEABLER(void *s,const void *e){return WRITEABLER(reinterpret_cast<BYTE *>(s),reinterpret_cast<const BYTE *>(e));}

// templates didn't work for this, the pointer types are too complex
#define EXECABLE(POINTER) (!IsBadCodePtr(reinterpret_cast<FARPROC>(POINTER)))

template<typename t>inline bool INRANGE(t v,t l,t h){return v>=l&&v<=h;}
template<typename t>inline bool INRANGEN(t v,t l,unsigned n){return INRANGE(v,l,l+n);}
template<typename t>inline bool BETWEEN(t v,t l,t h){return v>l&&v<h;}

#define REQUIRE_PTR(POINTER,TEST) ASSERT_POINTER_TEST("REQUIRE_PTR",TEST,#TEST,POINTER,#POINTER)
#define ENSURE_PTR(POINTER,TEST) ASSERT_POINTER_TEST("ENSURE_PTR",TEST,#TEST,POINTER,#POINTER)
#define ASSERT_PTR(POINTER,TEST) ASSERT_POINTER_TEST("ASSERT_PTR",TEST,#TEST,POINTER,#POINTER)
