#include "module.h"
#include "modtype.h"
#include "configstruct.h"
#include "droptarget.h"

// Set these values in modtype.h

const char * const ModuleName     = STRV(MODULE_NAME_LONG);
const char * const ModuleVersion  = STRV(MODULE_VERSION);
const char * const ModuleAuthor   = STRV(MODULE_AUTHOR);

/* array so sizeof() works */ 
const char ModuleBangPrefix[]     = "!" STRV(MODULE_NAME_SHORT);
const int ModuleBangPrefixLen     = sizeof(ModuleBangPrefix)-1;
const char ModuleRCPrefix[]       = STRV(MODULE_NAME_SHORT);
const int ModuleRCPrefixLen       = sizeof(ModuleRCPrefix)-1;
#ifndef SINGLE_WINDOW
const char ModuleListKey[]        = "*" STRV(MODULE_NAME_SHORT);
const int ModuleListKeyLen        = sizeof(ModuleListKey)-1;
#endif

#ifndef NDEBUG
DEFINE_MODULE_LOGGER(ModuleRCPrefix, STRV(MODULE_NAME_NLM));
#endif

//

#ifdef SINGLE_WINDOW
#define CONFIG_SW "Sngl "
#else
#define CONFIG_SW "Mlti "
#endif

#ifdef NO_SCRIPT_ACTIONS
#define CONFIG_SA "NoSc "
#else
#define CONFIG_SA "Scpt "
#endif

#ifdef USE_LSTOOLS
#define CONFIG_LS "LSTl "
#else
#define CONFIG_LS "StLS "
#endif

#ifdef NO_SKIN
#define CONFIG_SKIN "Whit "
#else
#define CONFIG_SKIN "Skin "
#endif

const char * const ModuleBuild = CONFIG_LS CONFIG_SW CONFIG_SA CONFIG_SKIN;

extern "C" {
__declspec( dllexport ) int initModuleEx(HWND parent, HINSTANCE dll, LPCSTR szPath);
__declspec( dllexport ) int quitModule(HINSTANCE dll);
#ifdef SINGLE_WINDOW
__declspec( dllexport ) int initWharfModule(HWND ParentWnd, HINSTANCE dllInst, wharfDataType *wd);
__declspec( dllexport ) int quitWharfModule(HINSTANCE dllInst);
#endif
}

LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);

ModInfo Module;

BOOL  (WINAPI *dAlphaBlend)( IN HDC, IN int, IN int, IN int, IN int,
    IN HDC, IN int, IN int, IN int, IN int, IN BLENDFUNCTION) = NULL;
BOOL (WINAPI *dUpdateLayeredWindow)(HWND,HDC,POINT*,SIZE*,HDC,POINT*,
		COLORREF,BLENDFUNCTION*,DWORD) = NULL;

#define LINKLIB(name) { name, NULL },
#define LINKFUNC(name,ptr) { name, reinterpret_cast<int**>(&ptr) },

struct lsLinkItem {
	const char *name;		// the name of the function/library
	int **ptr;					// the location to store its address or NULL for lib name
};

void LSLinkFunctions(lsLinkItem *list);
void LSUnlinkFunctions(lsLinkItem *list);

lsLinkItem DynFuncs[] =
{
	LINKFUNC("UpdateLayeredWindow",dUpdateLayeredWindow)

	LINKLIB("MSIMG32.DLL")
	LINKFUNC("AlphaBlend",dAlphaBlend)

	/* end of list marker */ 
	{ NULL, NULL }
};

// Initialize Module
//
//  ParentWnd = litestep or wharf window
//  dll       = instance handle for this module
//  szPath    = path to litestep.exe
//
// return 0 on success (failure is more or less ignored)
//
int initModuleEx(HWND ParentWnd, HINSTANCE dll, LPCSTR szPath)
{
  TRACE("----- initModuleEx");
  TRACE(ModuleVersion);
#define TEST_ASSERT_FAIL false
  DCATCHALL(ASSERT(TEST_ASSERT_FAIL));

	Config c;

  /* start with a clean slate */ 
	ZeroMemory(&Module,sizeof(ModInfo));
	Module.hInstance = dll;
  /* load optionally linked functions */ 
	LSLinkFunctions(DynFuncs);

  /* register window class */ 
	WNDCLASS wc;
	ZeroMemory(&wc,sizeof(WNDCLASS));
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.lpfnWndProc = &WndProc;
	wc.hInstance = dll;
	wc.lpszClassName = ModuleName;
	wc.cbWndExtra = sizeof(WndInfo *);
	if (!RegisterClass(&wc)) 
	{
		MessageBox(ParentWnd,
			"Error registering window class.\n\nPerhaps the module did not unload properly?",
			ModuleName,MB_OK|MB_ICONERROR);
		return 1;
	}

  /* find the appropriate parent for "always on bottom" and such */ 
	if (ParentWnd == GetLitestepWnd())
		/* using the litestep window as a parent will get me dropped behind  */ 
		/* the desktop if i'm in "always on bottom" mode, so use the desktop */ 
		Module.hBottomParentWnd = NULL;
	else
		Module.hBottomParentWnd = ParentWnd;
	if (!Module.hBottomParentWnd)
		Module.hBottomParentWnd = FindWindow("DesktopBackgroundClass", NULL);
	if (!Module.hBottomParentWnd)
		Module.hBottomParentWnd = GetDesktopWindow();

#ifdef SINGLE_WINDOW
	Module.MainWnd.hParentWnd = Module.hBottomParentWnd;
#endif

  LSInitMultiMon();

  SkinBoxDropTarget::RegisterFormats();

  /* read module-wide configuration */ 
	ZeroMemory(&c,sizeof(Config));
  {
    ConfigState cs(&c,ModuleRCPrefix,sizeof(ModuleRCPrefix)-1);
    CallParser(&c,NULL,&Module,true);
  }

  /* register bang commands */ 
	LSRegisterBangList(ModuleBangPrefix,sizeof(ModuleBangPrefix)-1,Bangs);

#ifdef SINGLE_WINDOW
  /* read config for window and create it */ 
	if (!Module.bWaitForBox)
	{
		ReadConfigWnd(&Module.MainWnd,&c,ModuleRCPrefix,sizeof(ModuleRCPrefix)-1);
		if (!CreateWnd(&Module.MainWnd)) return 1;
	}
#else
  /* read predefined windows */ 
  LSParseStarList(ModuleListKey,ModuleListKeyLen,ParseBoxProc,
    &Module.pWnds,sizeof(WndInfo),(DWORD)&c);
  TRACE("Done parsing...");
  CreateWindows();
#endif

#ifndef NO_SCRIPT_ACTIONS
  ExecAction(NULL,OnInit);
#endif

	return 0;
}

// Quit Module
//
//  dll = instance handle for module
//
// return 0 on success (failure is more or less ignored)
//
int quitModule(HINSTANCE dll)
{
#ifndef NO_SCRIPT_ACTIONS
  ExecAction(NULL,OnQuit);
#endif

#ifdef SINGLE_WINDOW
  /* destroy the main window */ 
	DestroyWindow(Module.MainWnd.hWnd);
#else
  LSFreeStarList(FreeWndProc,Module.pWnds,NULL);
#endif

  /* unregister bang commands */ 
	LSRemoveBangList(ModuleBangPrefix,sizeof(ModuleBangPrefix)-1,Bangs);

  LSQuitMultiMon();

  /* unregister window class */ 
	UnregisterClass(ModuleName,Module.hInstance);
  /* unload any dlls we loaded for optional funcs */ 
	LSUnlinkFunctions(DynFuncs);

	return 0;
}

#ifdef SINGLE_WINDOW
// hooks for loading as a wharf module
// just pass the params to regular init/quit functions
int initWharfModule(HWND ParentWnd, HINSTANCE dllInst, wharfDataType *wd)
{ return initModuleEx (ParentWnd, dllInst, (wd) ? wd->lsPath : NULL); }
int quitWharfModule(HINSTANCE dllInst)
{ return quitModule(dllInst); }
#endif


//
// old lstools stuff
//

int LSStringEnum(const char *string,const char **list,int inc)
{
  REQUIRE_PTR(string, READABLE(string));
  REQUIRE_PTR(list, READABLE(list));

	const char **pList = list;
	int i=0;

	while (*pList)
  {
    ASSERT_PTR(pList, READABLEN(pList, inc));
    ASSERT_PTR(*pList, READABLE(*pList));

    if (lstrcmpi(string,*pList)==0) return i;
		i++; pList += inc/sizeof(char*);
	}

	return -1;
}


void LSLinkFunctions(lsLinkItem *list)
{
	HINSTANCE hDLL;
  int *ptr;

	ASSERT(list!=NULL);

	// most common module to import from, so don't require its name
	hDLL = GetModuleHandle("USER32.DLL");

	while (list->name!=NULL) {
		if (list->ptr==NULL) {
			// it's a library name

			// load it (always load so I know which need to be unloaded, all of them)
			hDLL = LoadLibrary(list->name);

#if !defined(NDEBUG) && defined(TRACE_FILE)
      {
        char buf[256];
        sprintf(buf,"DLL: %s = %08X",list->name,hDLL);
        TRACE(TRACE_DUMP,buf);
      }
#endif
		} else {
			// it's a function name

			ptr = reinterpret_cast<int *>(GetProcAddress(hDLL, list->name));
      if (ptr) *list->ptr = ptr;

#if !defined(NDEBUG) && defined(TRACE_FILE)
      {
        char buf[256];
        sprintf(buf,"Func: %s = %08X",list->name,*list->ptr);
        TRACE(TRACE_DUMP,buf);
      }
#endif
		}

		// next item
		++list;
	}
}

void LSUnlinkFunctions(lsLinkItem *list)
{
	ASSERT(list!=NULL);

	while (list->name!=NULL) {
		if (list->ptr==NULL) {
			// it's a library name

			// unload it
			FreeLibrary(GetModuleHandle(list->name));

		} else {
			// it's a function name
			// nothing to do
		}

		// next item
		++list;
	}
}


void LSRegisterBangList(const char *sPrefix, unsigned nLen, lsBangCmdDef *Bangs)
{
	lsBangCmdDef *pBang;
	char BangString[64];

	TRACE("RegisterBangs");

  lstrcpy(BangString, sPrefix);

	pBang = Bangs;
	while (pBang->Name != NULL) {
		lstrcpy(BangString+nLen,pBang->Name);
		AddBangCommand(BangString, pBang->Command);
		pBang++;
	}
}

void LSRemoveBangList(const char *sPrefix, unsigned nLen, lsBangCmdDef *Bangs)
{
	lsBangCmdDef *pBang;
	char BangString[64];

	TRACE("RemoveBangs");

  lstrcpy(BangString, sPrefix);

	pBang = Bangs;
	while (pBang->Name != NULL) {
		lstrcpy(BangString+nLen,pBang->Name);
		RemoveBangCommand(BangString);
		pBang++;
	}
}


#define min3(a,b,c) (((a)<(b))?(min(a,c)):(min(b,c)))

// tile a piece of the source over a piece of the destination
// with sl,st aligned to x,y of destination
BOOL LSTileBltAligned(
	HDC d, int dl, int dt, int dw, int dh,
	HDC s, int sl, int st, int sw, int sh,
	int x, int y,
	DWORD rop)
{
	int i,j,top;
	BOOL ret = TRUE;

	top = -sh-(dt-y)%sh;
	i   = -sw-(dl-x)%sw;

	for (i;i<dw;i+=sw) {
		for (j=top;j<dh;j+=sh) {
			ret &=
			BitBlt(
				d,max(dl,dl+i),max(dt,dt+j),
				  min3(sw,dw-i,sw+i),min3(sh,dh-j,sh+j),
				s,max(sl,sl-i),max(st,st-j),rop);
		}
	}
	return ret;
}

// tile a piece of the source over a piece of the destination
BOOL LSTileBlt(
	HDC d, int dl, int dt, int dw, int dh,
	HDC s, int sl, int st, int sw, int sh,
	DWORD rop)
{
#if(1)
	return LSTileBltAligned(d,dl,dt,dw,dh,s,sl,st,sw,sh,dl,dt,rop);
#else
	int i,j;
	BOOL ret = TRUE;
	for (i=0;i<dw;i+=sw) {
		for (j=0;j<dh;j+=sh) {
			ret &=
			BitBlt(
				d,dl+i,dt+j,
				min(sw,dw-i),min(sh,dh-j),
				s,sl,st,rop);
		}
	}
	return ret;
#endif
}

#ifndef NDEBUG
static int teb_lev = 0;
#endif

#define TRANSPARENT_COLOR RGB(255,0,255)

BOOL LSTransEdgeBlt(
	HDC d, const RECT *rDest,
	HDC s, lsEdgeMap *emSrc, BOOL bTransBlt)
{
	/* from my own scbar/RabidBitmap                      */ 
	/* varnames have been changed to protect the innocent */ 
	/* This could be done in one pass with 9 blts, but    */ 
	/* I believe there were problems with StretchBlt and  */ 
	/* bitmaps that were only a few pixels wide and       */ 
	/* stretched vertically.  It was always off by one.   */ 
	struct { int w,h; } center;
	struct { int l,t,r,b; } squish;
	int bw,bh;

	const int dl = rDest->left, dt = rDest->top;
	const int dw = rDest->right-rDest->left, dh = rDest->bottom-rDest->top;
	const int sl = 0, st = 0;
	const int sw = emSrc->Size.cx, sh = emSrc->Size.cy;
	const int bl = emSrc->Edge.left, bt = emSrc->Edge.top;
	const int br = emSrc->Edge.right, bb = emSrc->Edge.bottom;
	const bool tile = emSrc->Tile!=FALSE;

	HDC dc1,dc2;
	HBITMAP tempbm1,tempbm2;
	HBITMAP old1,old2;

	BOOL ret = TRUE;

	center.w = dw - bl - br;
	center.h = dh - bt - bb;

	bw = sw - bl - br;
	bh = sh - bt - bb;

	/* stretch horizontally */ 
	if (dw == sw) {
		/* skip the first stretch step */ 
		dc1 = s;
	} else {
		/* create temp dc & bitmap     */ 
		dc1 = CreateCompatibleDC(d);
		tempbm1 = CreateCompatibleBitmap(d,dw,sh);
		old1 = (HBITMAP)SelectObject( dc1, tempbm1 );

		if (center.w < 0) {
			/* squeeze borders           */ 
			squish.l = dw*bl/(bl+br);
			squish.r = dw-squish.l;
			center.w = 0;
		} else {
			/* stretch/squeeze center    */ 
			squish.l = bl;
			squish.r = br;
		}
		if (squish.l)
			ret &= BitBlt(
				dc1,0,0,squish.l,sh,
				s,0,0,
				SRCCOPY);
		if (center.w) {
			if (tile)
				ret &= LSTileBlt(
					dc1,squish.l,0,center.w,sh,
					s,squish.l,0,bw,sh,
					SRCCOPY);
			else
				ret &= StretchBlt(
					dc1,squish.l,0,center.w,sh,
					s,squish.l,0,bw,sh,
					SRCCOPY);
		}
		if (squish.r)
			ret &= BitBlt(
				dc1,dw-squish.r,0,squish.r,sh,
				s,sw-squish.r,0,
				SRCCOPY);
	}

	/* stretch vertically */ 
	if (dh == sh) {
		/* skip second stretch step */ 
		dc2 = dc1;
	} else {
		/* create temp dc & bitmap  */ 
		dc2 = CreateCompatibleDC(d);
		tempbm2 = CreateCompatibleBitmap(d,dw,dh);
		old2 = (HBITMAP)SelectObject( dc2, tempbm2 );

		if (center.h < 0) {
			/* squeeze borders        */ 
			squish.t = dh*bt/(bt+bb);
			squish.b = dh-squish.t;
			center.h = 0;
		} else {
			/* stretch/squeeze center */ 
			squish.t = bt;
			squish.b = bb;
		}
		if (squish.t)
			ret &= BitBlt(
				dc2,0,0,dw,squish.t,
				dc1,0,0,
				SRCCOPY);
		if (center.h) {
			if (tile)
				ret &= LSTileBlt(
					dc2,0,squish.t,dw,center.h,
					dc1,0,squish.t,dw,bh,
					SRCCOPY);
			else
				ret &= StretchBlt(
					dc2,0,squish.t,dw,center.h,
					dc1,0,squish.t,dw,bh,
					SRCCOPY);
		}
		if (squish.b)
			ret &= BitBlt(
				dc2,0,dh-squish.b,dw,squish.b,
				dc1,0,sh-squish.b,
				SRCCOPY);
	}

	/* draw overlay */ 
	if (bTransBlt)
		TransparentBltLS(
			d,dl,dt,dw,dh,
			dc2,0,0,
			TRANSPARENT_COLOR);
	else
		BitBlt(
			d,dl,dt,dw,dh,
			dc2,0,0,
			SRCCOPY);

	/* free things  */ 
	if (dc1 != s) {
		SelectObject(dc1, old1);
		DeleteObject(tempbm1);
		DeleteDC(dc1);
	}
	if (dc2 != dc1) {
		SelectObject(dc2, old2);
		DeleteObject(tempbm2);
		DeleteDC(dc2);
	}

	return ret;
}


//
//  Desktop Bitmap Things
//

HDC hdcDesktop = 0;
HBITMAP hbmDesktop = 0;
DesktopPainterProc LSPaintDesktopHook = NULL;

// PaintDesktopEx and RegQueryStringValue were stolen from Maduin's label module.
// This version has been mutilated to fit my needs, so any bugs are my fault.

LPTSTR RegQueryStringValue(HKEY hKey, LPCTSTR pszSubKey, LPCTSTR pszValueName, LPTSTR pszBuffer, int nBufferSize, LPTSTR pszDefaultValue);

BOOL GetSmallestScreenSizeProc(LSMONITOR Mon,LPARAM Size)
{
	RECT rScreen;

	LSGetMonitorScreenRect(Mon,&rScreen);

	if (rScreen.right-rScreen.left<reinterpret_cast<SIZE *>(Size)->cx
		&&rScreen.bottom-rScreen.top<reinterpret_cast<SIZE *>(Size)->cy)
	{
		reinterpret_cast<SIZE *>(Size)->cx = rScreen.right-rScreen.left;
		reinterpret_cast<SIZE *>(Size)->cy = rScreen.bottom-rScreen.top;
	}

	return TRUE;
}

struct PDCData
{
	HDC hdcDesktop;
	HDC hdcWallpaper;
	SIZE sWallpaper;
	RECT rDesktop;
};

BOOL PaintDesktopCenterredProc(LSMONITOR Mon,LPARAM Data)
{
	RECT rScreen;
	RECT rSource; /* right&bottom are width&height */ 
	POINT pos;

	LSGetMonitorScreenRect(Mon,&rScreen);

	/* alight top-left monitor to 0,0 of bitmap */ 
	OffsetRect(&rScreen,
		-reinterpret_cast<PDCData *>(Data)->rDesktop.left,
		-reinterpret_cast<PDCData *>(Data)->rDesktop.top);

	pos.x = (rScreen.right+rScreen.left - reinterpret_cast<PDCData *>(Data)->sWallpaper.cx)/2;
	pos.y = (rScreen.bottom+rScreen.top - reinterpret_cast<PDCData *>(Data)->sWallpaper.cy)/2;

	/* check size in x direction */ 
	if (pos.x<rScreen.left)
	{
		/* crop it to the size of the screen */ 
		rSource.left = rScreen.left-pos.x;
		rSource.right = rScreen.right-rScreen.left;
		pos.x = rScreen.left;
	}
	else
	{
		/* no need to crop it */ 
		rSource.left = 0;
		rSource.right = reinterpret_cast<PDCData *>(Data)->sWallpaper.cx;
	}

	/* check size in y direction */ 
	if (pos.y<rScreen.top)
	{
		/* crop it to the size of the screen */ 
		rSource.top = rScreen.top-pos.y;
		rSource.bottom = rScreen.bottom-rScreen.top;
		pos.y = rScreen.top;
	}
	else
	{
		/* no need to crop it */ 
		rSource.top = 0;
		rSource.bottom = reinterpret_cast<PDCData *>(Data)->sWallpaper.cy;
	}

	/* paint it */ 
	BitBlt(
		reinterpret_cast<PDCData *>(Data)->hdcDesktop,
		pos.x,pos.y,
		rSource.right,rSource.bottom,
		reinterpret_cast<PDCData *>(Data)->hdcWallpaper,
		rSource.left,rSource.top,
		SRCCOPY);

	return TRUE;
}

/*
Load the background bitmap and paint it the way Windows would.
Some of these seem pretty brain dead, but there's nothing I can
do about it.

This takes a noticable amount of time, especially with larger
desktop area, so don't call it very often at all.

On return phdc will have a device context with the bitmap selected
into it, and phbm will have the bitmap to select into the device
contect to get the desktop bitmap back out.
*/

void LoadDesktopBitmap(HDC *phdc,HBITMAP *phbm)
{
	PDCData d;
	SIZE sScreen;

	HBITMAP hbmWallpaper;
	HBITMAP hbmDesktop;
	HDC hdcScreen;

	hdcScreen = GetDC(NULL);

	d.hdcDesktop = CreateCompatibleDC(hdcScreen);

	LSGetMonitorScreenRect(LSMONITOR_ALL,&d.rDesktop);
	int cxDesktop = d.rDesktop.right-d.rDesktop.left;
	int cyDesktop = d.rDesktop.bottom-d.rDesktop.top;

	hbmDesktop = CreateCompatibleBitmap(hdcScreen, cxDesktop, cyDesktop);
	hbmDesktop = (HBITMAP) SelectObject(d.hdcDesktop, hbmDesktop);

	RECT r;
	SetRect(&r, 0, 0, cxDesktop, cyDesktop);
	FillRect(d.hdcDesktop, &r, GetSysColorBrush(COLOR_DESKTOP));

	if (!EXECABLE(LSPaintDesktopHook)) LSPaintDesktopHook = NULL;

	if (LSPaintDesktopHook==NULL || LSPaintDesktopHook(d.hdcDesktop)==FALSE)
	{
		char wallpaper[MAX_PATH] = { 0 };

		RegQueryStringValue(HKEY_CURRENT_USER,
			"Control Panel\\Desktop", "Wallpaper", wallpaper,
			MAX_PATH, "");

		if(wallpaper[0])
		{
			int wallpaperStyle;
			int tileWallpaper;

			char buffer[32];

			RegQueryStringValue(HKEY_CURRENT_USER, "Control Panel\\Desktop", "WallpaperStyle", buffer, 32, "0");
			wallpaperStyle = atoi(buffer);

			RegQueryStringValue(HKEY_CURRENT_USER, "Control Panel\\Desktop", "TileWallpaper", buffer, 32, "0");
			tileWallpaper = atoi(buffer);

			hbmWallpaper = (HBITMAP) LoadImage(0, wallpaper, IMAGE_BITMAP, 0, 0, LR_LOADFROMFILE);

			if(hbmWallpaper)
			{
				BITMAP bm;
				GetObject(hbmWallpaper, sizeof(BITMAP), &bm);

				d.sWallpaper.cx = bm.bmWidth;
				d.sWallpaper.cy = bm.bmHeight;

				d.hdcWallpaper = CreateCompatibleDC(hdcScreen);
				hbmWallpaper = (HBITMAP) SelectObject(d.hdcWallpaper, hbmWallpaper);

				if(tileWallpaper)
				{
/*	Tile behavior for multiple monitors:

			Bitmap is tiled across all monitors,
			aligned to upper-left of primary (ie, 0,0)
*/

					LSTileBltAligned(d.hdcDesktop,
						0, 0,
						cxDesktop, cyDesktop,
						d.hdcWallpaper,
						0, 0,
						d.sWallpaper.cx, d.sWallpaper.cy,
						-d.rDesktop.left,-d.rDesktop.top,
						SRCCOPY);
				}
				else
				{
					if(wallpaperStyle == 2) /* stretch */ 
					{
/*	Stretch behavior for multiple monitors:

			Bitmap is stretched/shrunk to fit smallest
			monitor then centered to each.
*/

						sScreen.cx = sScreen.cy = 2147483647L;
						LSForEachMonitor(&GetSmallestScreenSizeProc,reinterpret_cast<LPARAM>(&sScreen));

						/* should always be true, but just in case */ 
						if (sScreen.cx<2147483647L)
						{
							HDC hdcStretch = CreateCompatibleDC(hdcScreen);
							HBITMAP hbmStretch = CreateCompatibleBitmap(hdcScreen,sScreen.cx,sScreen.cy);
							hbmStretch = (HBITMAP)SelectObject(hdcStretch,hbmStretch);

							SetStretchBltMode(hdcDesktop, STRETCH_DELETESCANS);

							StretchBlt(hdcStretch,
								0, 0,
								sScreen.cx, sScreen.cy,
								d.hdcWallpaper,
								0, 0,
								d.sWallpaper.cx, d.sWallpaper.cy,
								SRCCOPY);

							/* exchange it with the wallpaper */ 
							hbmWallpaper = (HBITMAP) SelectObject(d.hdcWallpaper, hbmWallpaper);
							DeleteDC(d.hdcWallpaper);
							DeleteObject(hbmWallpaper);

							d.hdcWallpaper = hdcStretch;
							hbmWallpaper = hbmStretch;
							d.sWallpaper = sScreen;

							/* then fall into the center code */ 
						}
					}

/*	Center behavior for multiple monitors:

			Bitmap is cropped to each monitor, centered
			with edges that don't fit simply cropped off.
			(Did I mention there's cropping involved?)
*/

					LSForEachMonitor(&PaintDesktopCenterredProc,reinterpret_cast<LPARAM>(&d));
				}

				hbmWallpaper = (HBITMAP) SelectObject(d.hdcWallpaper, hbmWallpaper);
				DeleteDC(d.hdcWallpaper);

				DeleteObject(hbmWallpaper);
			}
		}
	}

	ReleaseDC(0, hdcScreen);

	*phdc = d.hdcDesktop;
	*phbm = hbmDesktop;
}

BOOL LSSetPaintDesktopHook(DesktopPainterProc UserProc)
{
	if (!EXECABLE(LSPaintDesktopHook)) LSPaintDesktopHook = NULL;

	// only one at a time because I don't want to deal
	// with out of order hooking and unhooking
	if (LSPaintDesktopHook!=NULL && UserProc!=NULL) return FALSE;

	LSPaintDesktopHook = UserProc;

	return TRUE;
}

void LSFreeDesktopBmp(void)
{
	if (hbmDesktop && hdcDesktop)
	{
		hbmDesktop = (HBITMAP)SelectObject(hdcDesktop,hbmDesktop);
		DeleteObject(hbmDesktop);
		DeleteDC(hdcDesktop);
		hbmDesktop = NULL;
		hdcDesktop = NULL;
	}
}

// retrieve a string value from the registry
// seems a bit inefficient to do this for each string,
// but i'm lazy so I'll let Maduin's code do it.
LPTSTR RegQueryStringValue(HKEY hKey, LPCTSTR pszSubKey, LPCTSTR pszValueName, LPTSTR pszBuffer, int nBufferSize, LPTSTR pszDefaultValue)
{
	DWORD dwDataType;
	DWORD dwDataSize = (DWORD) nBufferSize;
	HKEY hSubKey = NULL;
	LONG lErrorCode;
	
	if(pszSubKey)
	{
		lErrorCode = RegOpenKeyEx(hKey, pszSubKey, 0, KEY_QUERY_VALUE, &hSubKey);
		
		if(lErrorCode != ERROR_SUCCESS)
		{
			lstrcpyn(pszBuffer, pszDefaultValue, nBufferSize);
			return pszBuffer;
		}
		
		hKey = hSubKey;
	}
	
	lErrorCode = RegQueryValueEx(hKey, pszValueName, NULL, &dwDataType,
		(LPBYTE) pszBuffer, &dwDataSize);
	
	if(hSubKey)
		RegCloseKey(hSubKey);
		
	if(dwDataType != REG_SZ || lErrorCode != ERROR_SUCCESS)
		lstrcpyn(pszBuffer, pszDefaultValue, nBufferSize);
	
	return pszBuffer;
}

// modelled after Maduin's PaintDesktopEx
void LSPaintDesktop(HDC hdcD, int xD, int yD, int cxD, int cyD,
										          int xS, int yS, BOOL bUpdateCache)
{
	if (hdcDesktop==NULL || hbmDesktop==NULL || bUpdateCache)
	{
		LSFreeDesktopBmp();
		LoadDesktopBitmap(&hdcDesktop,&hbmDesktop);
	}

	if(hdcD)
	{
		RECT rDesktop;
		LSGetMonitorScreenRect(LSMONITOR_ALL,&rDesktop);
		BitBlt(hdcD, xD, yD, cxD, cyD,
			hdcDesktop, xS-rDesktop.left, yS-rDesktop.top, SRCCOPY);
	}
}



