#include "module.h"
#include "configstruct.h"
#include "parsers.h"

const char * const s_EdgeColors::StyleEnum[] = {
  "flat", "raised", "sunken",
  "ridge", "groove",
  "raisedbutton", "sunkenbutton",
  NULL
};

ConfigItem PositionConfigItems[] =
{
  OPTCONF_ITEM("OnMonitor",   9,  StrToken,     s_Position,MonitorName, NULL,   NULL)
  OPTCONF_ITEM("X",           1,  Ordinate,     s_Position,X,           0,      0)
  OPTCONF_ITEM("Y",           1,  Ordinate,     s_Position,Y,           0,      0)
  OPTCONF_ITEM("Width",       5,  Ordinate,     s_Position,Width,       100,    0)
  OPTCONF_ITEM("Height",      6,  Ordinate,     s_Position,Height,      100,    0)
  { 0 }
};

ConfigItem WndFlagsConfigItems[] =
{
  CONF_FLAG('H',"Hidden",     6,                WndInfo,bHidden     )
  CONF_FLAG('L',"TrueAlpha",  9,                WndInfo,bTrueAlpha  )
  CONF_FLAG('T',"OnTop",      5,                WndInfo,bOnTop      )
  { 0 }
};

ConfigItem RectConfigItems[] =
{
  REQCONF_ITEM("Left",        4,  Long,         RECT,left,               0,     NULL)
  OPTCONF_ITEM("Top",         3,  Long,         RECT,top,               -1,     NULL)
  OPTCONF_ITEM("Right",       5,  Long,         RECT,right,             -1,     NULL)
  OPTCONF_ITEM("Bottom",      6,  Long,         RECT,bottom,            -1,     NULL)
  { 0 }
};

ConfigItem SkinConfigItems[] =
{
  REQCONF_ITEM("Bitmap",      6,  Bitmap,       lsEdgeMap,Bitmap,       NULL,                     NULL)
  OPTCONF_ITEM("Edge",        4,  UnitCompound, lsEdgeMap,Edge,         sizeof(RECT),             RectConfigItems)
  OPTCONF_ITEM("Tile",        4,  Bool,         lsEdgeMap,Tile,         true,                     NULL)
  { 0 }
};

ConfigItem ColorsConfigItems[] = 
{
  OPTCONF_ITEM("Style",   5,  Enum,   s_EdgeColors,Style,s_EdgeColors::RaisedButtonStyle,s_EdgeColors::StyleEnum)
  OPTCONF_ITEM("Face",    4,  Color,  s_EdgeColors,Face,      0xC0C0C0FF,         COLOR_3DFACE)
  OPTCONF_ITEM("Light",   5,  Color,  s_EdgeColors,Edge[0],   0xFFFFFFFF,         COLOR_3DHIGHLIGHT)
  OPTCONF_ITEM("Dark",    4,  Color,  s_EdgeColors,Edge[1],   0x808080FF,         COLOR_3DSHADOW)
  OPTCONF_ITEM("BrLight", 10, Color,  s_EdgeColors,Edge[2],   0xC0C0C0FF,         COLOR_3DFACE)
  OPTCONF_ITEM("DkDark",  9,  Color,  s_EdgeColors,Edge[3],   0xC0C0C0FF,         COLOR_3DDKSHADOW)
  { 0 }
};

ConfigItem WndConfigItems[] =
{
  OPTCONF_ITEM("Pos",         0,  Compound,     WndInfo,Pos,            sizeof(s_Position),       PositionConfigItems)
  OPTCONF_ITEM("Group",       0,  Flags,        WndInfo,Group,          0,                        WndFlagsConfigItems)
  OPTCONF_ITEM("Alpha",       5,  SimpleOrd,    WndInfo,Alpha,          255,                      255 )
  OPTCONF_ITEM("HoverTime",   9,  Long,         WndInfo,HoverTime,      0,                        NULL)
  OPTCONF_ITEM("On",          2,  ActionSet,    WndInfo,Actions,        ScriptActionWindowCount,  ScriptActionStrings)
  OPTCONF_ITEM("Skin",        0,  CompoundFree, WndInfo,emSkin,         sizeof(lsEdgeMap),        SkinConfigItems)
  OPTCONF_ITEM("Color",       5,  Compound,     WndInfo,cSkin,          sizeof(s_EdgeColors),     ColorsConfigItems)
  { 0 }
};

ConfigItem ModConfigItems[] =
{
  OPTCONF_ITEM("On",          2,  ActionSet,    ModInfo,Actions,        ScriptActionModuleCount,  ScriptActionStrings)
  OPTCONF_ITEM("HoverTime",   9,  Long,         ModInfo,HoverTime,      0,                        NULL)
  { 0 }
};

ConfigItem ConfigRootTable[] =
{
  CONF_ROOT(CompoundFree,   sizeof(ModInfo),        ModConfigItems)
  CONF_ROOT(CompoundFree,   sizeof(WndInfo),        WndConfigItems)
  CONF_ROOT(Compound,       sizeof(s_Position),     PositionConfigItems)
  CONF_ROOT(UnitCompound,   sizeof(RECT),           RectConfigItems)
  OPTCONF_ITEM(0,0, Bool, WndInfo,bHidden,false,0)
  CONF_ROOT(Compound,       sizeof(s_Position),     &PositionConfigItems[3])
  CONF_ROOT(Compound,       sizeof(s_EdgeColors),   ColorsConfigItems)
  { 0 }
};

const ConfigItem * const WndOnTop_ConfigItemPtr = &WndFlagsConfigItems[2];
const ConfigItem * const SkinTile_ConfigItemPtr = &SkinConfigItems[2];
const ConfigItem * const Alpha_ConfigItemPtr = &WndConfigItems[2];

const ConfigItem * const ModInfo     ::ConfigItemPtr = &ConfigRootTable[0];
const ConfigItem * const WndInfo     ::ConfigItemPtr = &ConfigRootTable[1];
const ConfigItem * const s_Position  ::ConfigItemPtr = &ConfigRootTable[2];
const ConfigItem * const          RECT_ConfigItemPtr = &ConfigRootTable[3];
const ConfigItem * const       WndShow_ConfigItemPtr = &ConfigRootTable[4];
const ConfigItem * const          Size_ConfigItemPtr = &ConfigRootTable[5];
const ConfigItem * const s_EdgeColors::ConfigItemPtr = &ConfigRootTable[6];



bool ParseBoxProc(void *elem,const char *line,DWORD conf)
{
	WndInfo *w = (WndInfo *)elem;
	Config *c = (Config *)conf;
	ASSERT_WRITEABLE(w);
	ASSERT_WRITEABLE(c);
	ASSERT_READABLE(line);
	char token[MAX_LINE_LENGTH];

  TRACE(line);

	if (!GetToken(line,token,&line,FALSE))
    return false;

	w->Name = new char[lstrlen(token)+1];
	lstrcpy(w->Name,token);
	w->hParentWnd = Module.hBottomParentWnd;

  {
    RECT rMonitor;
    ConfigState cs(c,token,lstrlen(token));
    CallParser(c,line,w,true);

    if (w->emSkin.Edge.top == -1) w->emSkin.Edge.top = w->emSkin.Edge.left;
    if (w->emSkin.Edge.right == -1) w->emSkin.Edge.right = w->emSkin.Edge.left;
    if (w->emSkin.Edge.bottom == -1) w->emSkin.Edge.bottom = w->emSkin.Edge.top;

    TRACE(w->Pos.MonitorName);
    w->Pos.Monitor = LSGetMonitor(w->Pos.MonitorName);
	  if (w->hParentWnd != Module.hBottomParentWnd)
	  {
      TRACE("Parent rect");
		  GetWindowRect(w->hParentWnd,&rMonitor);
    }
	  else
	  {
		  LSGetMonitorScreenRect(w->Pos.Monitor,&rMonitor);
	  }
    w->szParent.cx = rMonitor.right - rMonitor.left;
    w->szParent.cy = rMonitor.bottom - rMonitor.top;
    CalcOrd(&w->Pos.X,0,w->szParent.cx);
    CalcOrd(&w->Pos.Y,0,w->szParent.cy);
    CalcOrd(&w->Pos.Width,0,w->szParent.cx);
    CalcOrd(&w->Pos.Height,0,w->szParent.cy);

    CalcColorSet(&w->cSkin);
  }

	return true;
}

void FreeWndProc(void *elem,DWORD)
{
	WndInfo *w=(WndInfo*)elem;
	ASSERT_WRITEABLE(w);
	DestroyWindow(w->hWnd);

  CallFreer(w);
}





